<?xml version='1.0' encoding='UTF-8'?>
<!DOCTYPE gui SYSTEM 'kpartgui.dtd'>
<gui version="1" name="kajongg" >
 <MenuBar>
  <Menu name="kajongggame" >
   <text context="kajongg" >Game</text>
   <Action name="scoreGame" />
   <Action name="play" />
   <Action name="abort" />
   <Action name="quit" />
  </Menu>
  <Menu name="kajonggview" >
   <Text>View</Text>
   <Action name="scoring" />
   <Action name="scoreTable" />
   <Action name="explain" />
   <Action name="chat" />
  </Menu>
  <Menu name="settings" >
   <Action name="players" />
   <Action name="rulesets" />
   <Action name="angle" />
   <Action name="fullscreen" />
   <Action name="demoMode" />
  </Menu>
 </MenuBar>
 <ToolBar noMerge="1" name="mainToolBar" >
  <text>Toolbar</text>
  <Action name="quit" />
  <Action name="load" />
  <Action name="play" />
  <Action name="scoreTable" />
  <Action name="explain" />
  <Action name="players" />
 </ToolBar>
</gui>
