#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from server import kajonggServer
from util import checkMemory

kajonggServer()
checkMemory()
# profileMe()
